use crate::{Action, Context, Error, RecoverableError, Required};
use activitystreams::{base::AnyBase, prelude::*};

mod block;
mod follow;
mod like;
mod note;
mod person;

use block::{block, undo_block};
use follow::{accept_follow, follow, reject_follow, undo_accept_follow, undo_follow};
use like::{like, undo_like};
use note::{announce_note, create_note, delete_note, note, update_note};
use person::{create_person, delete_person, person, update_person};

pub(crate) fn ingest(
    any_base: AnyBase,
    ctx: &Context,
    mut stack: Vec<AnyBase>,
) -> Result<(), Error> {
    // Don't double-ingest
    if ctx.apub.object(any_base.id().req()?)?.is_some() {
        return Ok(());
    }

    match parse(any_base.clone(), ctx)? {
        Ok(action) => {
            if let Some(outbound) = action.perform(ctx)? {
                ctx.deliver(&*outbound);
            }
            if let Some(any_base) = stack.pop() {
                ctx.spawner.process(any_base, stack);
            }
        }
        Err(RecoverableError::MissingApub(url)) => {
            stack.push(any_base);
            ctx.spawner.download_apub(url, stack);
        }
        Err(RecoverableError::MissingImages(urls)) => {
            stack.push(any_base);
            ctx.spawner.download_images(urls, stack);
        }
    }

    Ok(())
}

fn get_obj<T, Kind>(
    any_base: AnyBase,
    ctx: &Context,
    f: impl FnOnce(T, &Context, AnyBase) -> Result<Result<Box<dyn Action>, RecoverableError>, Error>,
) -> Result<Result<Box<dyn Action>, RecoverableError>, Error>
where
    T: ActorAndObjectRefExt + ExtendsExt<Kind>,
    <T as activitystreams::base::Extends<Kind>>::Error: From<serde_json::Error>,
    Error: From<<T as activitystreams::base::Extends<Kind>>::Error>,
    for<'de> Kind: serde::Deserialize<'de>,
    Kind: serde::Serialize,
{
    wrap(any_base, ctx, |activity: T, ctx| {
        let object = activity.object().as_one().req()?;
        let any_base = if object.kind_str().is_none() {
            let id = object.id().req()?;
            if ctx.apub.deleted(id)? {
                return Err(Error::Invalid);
            }
            match ctx.apub.object(id)? {
                Some(any_base) => any_base,
                None => return Ok(Err(RecoverableError::MissingApub(id.to_owned()))),
            }
        } else {
            object.clone()
        };

        (f)(activity, ctx, any_base.clone())
    })
}

fn wrap<T, Kind>(
    any_base: AnyBase,
    ctx: &Context,
    f: impl FnOnce(T, &Context) -> Result<Result<Box<dyn Action>, RecoverableError>, Error>,
) -> Result<Result<Box<dyn Action>, RecoverableError>, Error>
where
    T: ExtendsExt<Kind>,
    <T as activitystreams::base::Extends<Kind>>::Error: From<serde_json::Error>,
    Error: From<<T as activitystreams::base::Extends<Kind>>::Error>,
    for<'de> Kind: serde::Deserialize<'de>,
    Kind: serde::Serialize,
{
    let activity: T = any_base.clone().extend()?.req()?;

    let res = (f)(activity, ctx);
    if res.as_ref().map(|r| r.is_ok()).unwrap_or(false) {
        ctx.apub.store_object(&any_base)?;
    }
    res
}

fn parse(
    any_base: AnyBase,
    ctx: &Context,
) -> Result<Result<Box<dyn Action>, RecoverableError>, Error> {
    match any_base.kind_str().req()? {
        "Accept" => get_obj(any_base, ctx, |accept, ctx, any_base| {
            match any_base.kind_str().req()? {
                "Follow" => wrap(any_base, ctx, |follow, ctx| {
                    accept_follow(&accept, &follow, ctx)
                }),
                _ => Err(Error::Invalid),
            }
        }),
        "Announce" => get_obj(any_base, ctx, |announce, ctx, any_base| {
            match any_base.kind_str().req()? {
                "Note" => wrap(any_base, ctx, |note, ctx| {
                    announce_note(&announce, &note, ctx)
                }),
                _ => Err(Error::Invalid),
            }
        }),
        "Block" => wrap(any_base, ctx, |block_obj, ctx| block(&block_obj, ctx)),
        "Create" => get_obj(any_base, ctx, |create, ctx, any_base| {
            match any_base.kind_str().req()? {
                "Note" => wrap(any_base, ctx, |note, ctx| create_note(&create, &note, ctx)),
                "Person" => wrap(any_base, ctx, |person, ctx| {
                    create_person(&create, &person, ctx)
                }),
                _ => Err(Error::Invalid),
            }
        }),
        "Delete" => get_obj(any_base, ctx, |delete, ctx, any_base| {
            match any_base.kind_str().req()? {
                "Note" => wrap(any_base, ctx, |note, ctx| delete_note(&delete, &note, ctx)),
                "Person" => wrap(any_base, ctx, |person, ctx| {
                    delete_person(&delete, &person, ctx)
                }),
                _ => Err(Error::Invalid),
            }
        }),
        "Follow" => wrap(any_base, ctx, |follow_obj, ctx| follow(&follow_obj, ctx)),
        "Like" => wrap(any_base, ctx, |like_obj, ctx| like(&like_obj, ctx)),
        "Note" => wrap(any_base, ctx, |note_obj, ctx| note(&note_obj, ctx)),
        "Person" => wrap(any_base, ctx, |person_obj, ctx| person(&person_obj, ctx)),
        "Reject" => get_obj(any_base, ctx, |reject, ctx, any_base| {
            match any_base.kind_str().req()? {
                "Follow" => wrap(any_base, ctx, |follow, ctx| {
                    reject_follow(&reject, &follow, ctx)
                }),
                _ => Err(Error::Invalid),
            }
        }),
        "Undo" => get_obj(any_base, ctx, |undo, ctx, any_base| {
            match any_base.kind_str().req()? {
                "Block" => wrap(any_base, ctx, |block, ctx| undo_block(&undo, &block, ctx)),
                "Accept" => get_obj(any_base, ctx, |accept, ctx, any_base| {
                    match any_base.kind_str().req()? {
                        "Follow" => wrap(any_base, ctx, |follow, ctx| {
                            undo_accept_follow(&undo, &accept, &follow, ctx)
                        }),
                        _ => Err(Error::Invalid),
                    }
                }),
                "Follow" => wrap(any_base, ctx, |follow, ctx| {
                    undo_follow(&undo, &follow, ctx)
                }),
                "Like" => wrap(any_base, ctx, |like, ctx| undo_like(&undo, &like, ctx)),
                _ => Err(Error::Invalid),
            }
        }),
        "Update" => get_obj(any_base, ctx, |update, ctx, any_base| {
            match any_base.kind_str().req()? {
                "Note" => wrap(any_base, ctx, |note, ctx| update_note(&update, &note, ctx)),
                "Person" => wrap(any_base, ctx, |person, ctx| {
                    update_person(&update, &person, ctx)
                }),
                _ => Err(Error::Invalid),
            }
        }),
        _ => Err(Error::Invalid),
    }
}
